-------------------------------------------------------------------------------
-- Localization libraries
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("SimpleSelfRebuff")
local BS = AceLibrary("Babble-Spell-2.2")

-------------------------------------------------------------------------------
-- Aliases & constants
-------------------------------------------------------------------------------

local TEH_BUTTON_NAME = "SimpleSelfRebuffButton"

local trackings = {}
do
	local function initTrackings(...)
		for i=1,select('#', ...) do
			local t = select(i, ...)			
			trackings[BS:GetSpellIcon(t)] = BS[t]
		end
	end
	initTrackings(
		'Track Beasts',     'Track Demons', 'Track Dragonkin',
		'Track Elementals', 'Track Giants', 'Track Hidden',
		'Track Humanoids',  'Track Undead', 'Find Herbs',
		'Find Minerals',    'Find Treasure'
	)
end

local activationCheckEvents = {
	'PLAYER_DEAD', 'PLAYER_UNGHOST', 'PLAYER_ALIVE',
	'PLAYER_UPDATE_RESTING',
	'PLAYER_REGEN_DISABLED', 'PLAYER_REGEN_ENABLED',
	'PLAYER_CONTROL_GAINED', 'PLAYER_CONTROL_LOST',
}

local CATEGORY_TRACKING = L["Tracking"]
local CATEGORY_MAINHAND = L["Main weapon"]
local CATEGORY_OFFHAND  = L["Off-hand weapon"]

local SOURCE_SPELL = 'spell'
local SOURCE_ITEM  = 'item'

local TARGET_AURA     = 'aura'
local TARGET_TRACKING = 'tracking'
local TARGET_MAINHAND = 'main-hand'
local TARGET_OFFHAND  = 'off-hand'

local SLOT_MAINHAND = GetInventorySlotInfo('MainHandSlot')
local SLOT_OFFHAND  = GetInventorySlotInfo('SecondaryHandSlot')

local invSlots = {
	[CATEGORY_MAINHAND] = SLOT_MAINHAND,
	[CATEGORY_OFFHAND]  = SLOT_OFFHAND,
	[TARGET_MAINHAND]   = SLOT_MAINHAND,
	[TARGET_OFFHAND]    = SLOT_OFFHAND,	
}

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local categories = {}
local actualBuffs = {}
local actualBuffExpirations = {}
local buffsPerTarget = {}

local buffToCast
local castButton

local knownSpells

local active

local db

local sources = {
	[SOURCE_SPELL] = {},
	[SOURCE_ITEM]  = {},
}

local targets = {
	[TARGET_AURA]      = {},
	[TARGET_TRACKING]  = {},
	[TARGET_MAINHAND]  = {},
}

-- Weapon buffs are both handled the same way
targets[TARGET_OFFHAND] = targets[TARGET_MAINHAND]

-------------------------------------------------------------------------------
-- Externals
-------------------------------------------------------------------------------

local Tablet = AceLibrary('Tablet-2.0')
 
-------------------------------------------------------------------------------
-- Addon declaration
-------------------------------------------------------------------------------

SimpleSelfRebuff = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "FuBarPlugin-2.0")
local SimpleSelfRebuff = SimpleSelfRebuff

SimpleSelfRebuff.options = {	
	type = 'group',
	args = {
		buffHeader = {
			type = 'header',
			name = ' ',
			order = 199,
		}
	}
}
	
do
	local self = SimpleSelfRebuff

	self.CATEGORY_TRACKING = CATEGORY_TRACKING
	self.CATEGORY_MAINHAND = CATEGORY_MAINHAND 
	self.CATEGORY_OFFHAND  = CATEGORY_OFFHAND 
	
	self.SOURCE_SPELL = SOURCE_SPELL
	self.SOURCE_ITEM  = SOURCE_ITEM
	
	self.TARGET_AURA     = TARGET_AURA
	self.TARGET_TRACKING = TARGET_TRACKING 
	self.TARGET_MAINHAND = TARGET_MAINHAND
	self.TARGET_OFFHAND  = TARGET_OFFHAND

	self.categories = categories
	self.buffsPerTarget = buffsPerTarget
	self.actualBuffs = actualBuffs
	self.actualBuffExpirations = actualBuffExpirations	
end

-------------------------------------------------------------------------------
-- FuBar plugin settings
-------------------------------------------------------------------------------

--SimpleSelfRebuff.hasIcon = true
SimpleSelfRebuff.hideWithoutStandby = true
SimpleSelfRebuff.OnMenuRequest = SimpleSelfRebuff.options

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function error(msg, ...)
	if select('#', ...) > 1 then
		msg = msg:format(...)
	end
	return _G.error(msg)
end

local function erase(...)
	for i=1,select('#', ...) do
		local t = select(i, ...)
		if type(t) == 'table' then
			for k in pairs(t) do
				t[k] = nil 
			end
		end
	end
end

local function colorize(text, r, g, b)
	if type(r) == 'table' and r.r then
		return colorize(msg, r.r, r.g, r.b)
	else
		return ("|cff%02x%02x%02x%s|r"):format(math.floor(r*255), math.floor(g*255), math.floor(b*255), text)
	end
end

-------------------------------------------------------------------------------
-- Initializing
-------------------------------------------------------------------------------

-- Addon setup
function SimpleSelfRebuff:OnInitialize()
	self.revision = tonumber(string.sub("$Revision: 42134 $", 12, -3))
	self.version = "r" .. SimpleSelfRebuff.revision
	self.date = string.sub("$Date: 2007-06-28 19:08:52 -0400 (Thu, 28 Jun 2007) $", 8, 17)

	self:RegisterDB("SimpleSelfRebuffDB")
	self:RegisterDefaults('profile', {
		categories = {},
	})

	db = self.db.profile

	self:RegisterChatCommand({"/ssr", "/simpleselfrebuff"}, self.options)
	
	if AceLibrary:HasInstance('AceDebug-2.0') then
		AceLibrary('AceDebug-2.0'):embed(SimpleSelfRebuff)
		--SimpleSelfRebuff:SetDebugging(UnitName('player') == 'Adirelle')
	else
		function SimpleSelfRebuff.Debug() end
	end
		
end

function SimpleSelfRebuff:OnEnable(first)
	if first then
		self:ProcessRegistry()
	end

	self:RegisterBucketEvent('SPELLS_CHANGED', 0.1, 'ScanSpellbook')
	self:RegisterBucketEvent(activationCheckEvents, 0.1, 'CheckActivation')
	self:RegisterBucketEvent('PLAYER_AURAS_CHANGED', 0.2)
	self:RegisterBucketEvent('SimpleSelfRebuff_UpdateDisplay', 0.1, 'UpdateDisplay')
	
	self:ScanSpellbook()
	self:CheckActivation()
end

function SimpleSelfRebuff:PLAYER_AURAS_CHANGED()
	self:CheckActivation()
	if active then
		self:ScanPlayerAuras()
	end
end

-------------------------------------------------------------------------------
-- Options handling
-------------------------------------------------------------------------------

local function getSelectCategoryBuff(category)
	return db.categories[category] or "-"
end

local function getToggleCategoryBuff(category)
	return db.categories[category]
end

local function setCategoryBuff(category, name)
	if name == '-' then
		name = nil
	end
	db.categories[category] = name
	SimpleSelfRebuff:CheckRebuff()
end

function SimpleSelfRebuff:BuildOptions()
	local opts = self.options.args
	for categoryName, category in pairs(categories) do
		local categoryName, category = categoryName, category
		if category.count > 1 then
			opts[categoryName] = {
				type = 'text',
				name = categoryName,
				desc = categoryName,
				passValue = categoryName,
				get = getSelectCategoryBuff,
				set = setCategoryBuff,
				validate = {},
				order = 200,
			}
		else
			opts[categoryName] = {
				type = 'toggle',
				name = categoryName,
				desc = categoryName,
				passValue = categoryName,
				get = getToggleCategoryBuff,
				set = setCategoryBuff,
				order = 200,
			}
		end
	end
	self:UpdateOptions()
end

function SimpleSelfRebuff:UpdateOptions()
	for categoryName, category in pairs(categories) do
		if category.count > 1 then
			local validate = self.options.args[categoryName].validate
			erase(validate)
			validate['-'] = L['None']
			for name,buff in ipairs(category.buffs) do
				if buff.recast and (buff.source == SOURCE_ITEM or knownSpells[name]) then
					validate[name] = name
				end
			end
		end
	end
end

-------------------------------------------------------------------------------
-- Buff registry
-------------------------------------------------------------------------------

do
	local buffFuncs = {}

	function SimpleSelfRebuff:RegisterBuffSetup(func)
		if type(func) == "function" then
			tinsert(buffFuncs, func)
		else
			error("Arg #1 to RegisterBuffs must be a function")
		end
	end
	
	function SimpleSelfRebuff:ProcessRegistry()
		if not buffFuncs then
			return
		end
		
		local selfClassName = select(2, UnitClass('player'))
		for i,func in ipairs(buffFuncs) do
			func(self, BS, selfClassName, L)
		end
		buffFuncs = nil

		self:BuildOptions()
	end
	
end

function SimpleSelfRebuff:RegisterCategory(name, ...)
	local category = categories[name]
	if not category then
		category = {
			name = name,
			count = 0,
			buffs = {},
		}
		if select('#', ...) > 0 then
			for i = 1, select('#', ...), 2 do
				local k,v = select(i, ...)
				category[k] = v
			end
		end	
		categories[name] = category
	end
	return category
end

function SimpleSelfRebuff:RegisterBuff(name, categoryName, recast, target, source, ...)

	categoryName = categoryName or name
	target = target or TARGET_AURA
	source = source or SOURCE_SPELL
	recast = (recast == nil or recast) and true
	
	local category = self:RegisterCategory(categoryName)
	if category.buffs[name] then
		error("%q already registered in category", name)
	end
	local buff = {
		name     = name,
		category = category,
		recast   = recast,
		target   = target,
		source   = source,
	}
	if select('#', ...) > 0 then
		for i = 1, select('#', ...), 2 do
			local k,v = select(i, ...)
			buff[k] = v
		end
	end
	category.buffs[name] = buff
	if buff.recast then
		category.recastableBuffs = true
	end
	category.count = category.count + 1
	if not buffsPerTarget[target] then
		buffsPerTarget[target] = {}
	end
	buffsPerTarget[target][name] = buff
	return buff
end

-------------------------------------------------------------------------------
-- Activation
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckActivation()
	local shouldActivate = not (UnitIsDeadOrGhost('player') or InCombatLockdown() or IsResting()
	  or IsMounted() or IsFlying() or IsStealthed())

	if shouldActivate and not active then
		active = true
		self:Debug('Activated')
		
		self:ScheduleRepeatingEvent('SimpleSelfRebuff_CheckRebuff', self.CheckRebuff, 1.5, self)
		self:ScanPlayerAuras()
		self:CheckRebuff()
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
		
	elseif not shouldActivate and active then
		active = nil
		self:Debug('Disactivated')
		
		erase(actualBuffs, actualBuffExpirations)

		self:CleanupPendingCast()		
		self:CancelScheduledEvent('SimpleSelfRebuff_CheckRebuff')
		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

-------------------------------------------------------------------------------
-- ActionButton setting per type of buff
-------------------------------------------------------------------------------

sources[SOURCE_SPELL].SetupBuffSource = function(self, castButton, buff)
	self:Debug('Setup casting for spell %q', buff.name)
	castButton:SetAttribute('type', 'spell')
	castButton:SetAttribute('spell', buff.name)
end

sources[SOURCE_ITEM].SetupBuffSource = function(self, castButton, buff)
	castButton:SetAttribute('type', 'item')
	castButton:SetAttribute('item', buff.name)
end

-------------------------------------------------------------------------------
-- ActionButton setting per target of buff
-------------------------------------------------------------------------------

targets[TARGET_AURA].SetupBuffTarget = function(self, castButton, buff)
	castButton:SetAttribute('unit', 'player')
end

targets[TARGET_MAINHAND].SetupBuffTarget = function(self, castButton, buff)
	castButton:SetAttribute('target-slot', invSlots[buff.target])
end

-- Tracking is cast as an aura target
targets[TARGET_TRACKING].SetupBuffTarget = targets[TARGET_AURA].SetupBuffTarget

-------------------------------------------------------------------------------
-- Core
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckRebuff()

	local newBuffToCast = nil	
	local shortestBuff, shortestBuffDuration
	
	for categoryName,category in pairs(categories) do
		local expectedBuff = category.buffs[db.categories[categoryName] or "-"]
		
		if expectedBuff then
			-- Always check cooldown
			local inCooldown
			local start, duration = GetSpellCooldown(expectedBuff.name)
			if start and start + duration > GetTime() then
				inCooldown = true
			end
			
			-- Is we can cast the spell, check it further
			if not inCooldown and IsUsableSpell(expectedBuff.name) then			
				local actualBuff = actualBuffs[categoryName] 
				
				-- Check if it is missing
				if not actualBuff then
					newBuffToCast = expectedBuff
					--self:Debug('Missing buff: %q', expectedBuff)
					break
					
				-- Check remaning time
				elseif actualBuff.name == expectedBuff.name and actualBuffExpirations[categoryName] then
					local timeLeft = actualBuffExpirations[categoryName] - GetTime()
					if timeLeft > 0 then
						if not shortestBuffDuration or timeLeft < shortestBuffDuration then
							shortestBuffDuration, shortestBuff = timeLeft, actualBuff
						end
					end
				end
				
			end
		end
	end
	
	if not newBuffToCast and shortestBuffName and shortestBuffDuration < 30 then
		newBuffToCast = shortestBuff
		self:Debug('Buff needs refresh: %q (%q)', newBuffToCast, shortestBuffDuration)
	end

	if newBuffToCast ~= buffToCast then
		
		if newBuffToCast then
			buffToCast = newBuffToCast
			
			if not castButton then
				castButton = CreateFrame("Button", TEH_BUTTON_NAME, UIParent, "SecureActionButtonTemplate")
				castButton:RegisterForClicks("LeftButtonUp")
				castButton:SetPoint("CENTER",0,0)
				castButton:Hide()
			end
			
			-- Reset attributes
			castButton:SetAttribute('type', nil)
			castButton:SetAttribute('spell', nil)			
			castButton:SetAttribute('unit', nil)			
			castButton:SetAttribute('target-slot', nil)			
			
			-- Setup attributes depending on source and target
			self:PrintLiteral(buffToCast)
			sources[buffToCast.source].SetupBuffSource(self, castButton, buffToCast)
			targets[buffToCast.target].SetupBuffTarget(self, castButton, buffToCast)
			
			-- Binding
			self:Debug('Binding mousewheel')
			SetOverrideBindingClick(castButton, true, "MOUSEWHEELUP", TEH_BUTTON_NAME, "LeftButton")
			SetOverrideBindingClick(castButton, true, "MOUSEWHEELDOWN", TEH_BUTTON_NAME, "LeftButton")	

		else
			self:CleanupPendingCast()
		end
		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end

end

function SimpleSelfRebuff:CleanupPendingCast()
	if buffToCast then
		self:Debug('Unbinding mousewheel')
		buffToCast = nil
		if castButton then
			ClearOverrideBindings(castButton)
		end
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

-------------------------------------------------------------------------------
-- Spellbook handling
-------------------------------------------------------------------------------

knownSpells = setmetatable({}, {
	__index = function(self, key) 
		local i = 1
		while true do
			local name = GetSpellName(i, BOOKTYPE_SPELL)
			if not name then
				break
			end
			rawset(self, name, true)
			i = i + 1
		end
		return rawget(self, key)
	end
})

function SimpleSelfRebuff:ScanSpellbook() 
	self:UpdateOptions()	
	self:CheckActivation()
end

-------------------------------------------------------------------------------
-- Player buff status tracking
-------------------------------------------------------------------------------

do
	local seen = {}
	local dirty
	
	local function seenBuff(buff, timeLeft)
		local category = buff.category
		seen[category.name] = true
		if not actualBuffs[category.name] then
			actualBuffs[category.name] = buff
			SimpleSelfRebuff:Debug('New buff: %q', buff.name)			
			dirty = true
		end
		if timeLeft and math.floor(timeLeft) > 0 then
			actualBuffExpirations[category.name] = GetTime() + timeLeft		
		else
			actualBuffExpirations[category.name] = nil
		end		
	end

	-- Buff scanning per type of buff
		
	targets[TARGET_AURA].Scan = function(self, target, buffs)
		for i = 1, 32 do
			local index = GetPlayerBuff(i, "HELPFUL")
			local buffName = index and GetPlayerBuffName(index)		
			local buff = buffName and buffs[buffName]
			if buff then
				seenBuff(buff, GetPlayerBuffTimeLeft(index))
			end
		end
	end
	
	targets[TARGET_TRACKING].Scan = function(self, target, buffs)
		local trackingTexture = GetTrackingTexture()
		local trackingName = trackingTexture and trackings[trackingTexture]
		local trackingBuff = trackingName and buffs[trackingName]
		if trackingBuff then
			seenBuff(trackingBuff)
		end
	end

	targets[TARGET_MAINHAND].Scan = function(self, target)
	end
	
	-- Main scanning method
	
	function SimpleSelfRebuff:ScanPlayerAuras()
		if next(seen) then erase(seen) end
		dirty = nil
		
		for target, buffs in pairs(buffsPerTarget) do
			targets[target].Scan(self, target, buffs)
		end
		
		for category, actualBuff in pairs(actualBuffs) do
			if not seen[category] then
				actualBuffs[category] = nil
				actualBuffExpirations[category] = nil
				self:Debug('Faded out: %q', actualBuff.name)
				dirty = true
			end
		end
		
		if dirty then
			dirty = nil
			self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
			self:Debug('Auras changed - checking rebuff')
			self:CheckRebuff()
		end
		erase(seen)
		
	end
	
end

-------------------------------------------------------------------------------
-- FuBar Plugin info
-------------------------------------------------------------------------------

local function GetBuffInfo(category)

	local expectedBuff = category.buffs[db.categories[category.name] or "-"]
	if not expectedBuff then
		return '-', '-', 0.5, 0.5, 0.5
	end
	
	local actualBuff = actualBuffs[category.name]
	if not actualBuff then
		return expectedBuff.name, L['Missing'], 1, 0, 0
	end
	
	local actualExpiration = actualBuffExpirations[category.name]
	local timeLeft
	if actualExpiration then
		timeLeft = math.floor(actualExpiration - GetTime())
		if timeLeft >= 3600 then
			timeLeft = HOUR_ONELETTER_ABBR:format(math.ceil(timeLeft/3600))
		elseif timeLeft >= 60 then
			timeLeft = MINUTE_ONELETTER_ABBR:format(math.ceil(timeLeft/60))
		else
			timeLeft = SECOND_ONELETTER_ABBR:format(timeLeft)
		end	
	else
		timeLeft = L['n/a']
	end
	local color
	if actualBuff.name == expectedBuff.name then
		return actualBuff.name, timeLeft, 0, 1, 0
	else
		return actualBuff.name, timeLeft, 1, 1, 0
	end
end

function SimpleSelfRebuff:OnTextUpdate()
	if active then
		if buffToCast then
			local name, timeLeft, r, g, b = GetBuffInfo(buffToCast.category)
			self:SetText(colorize(timeLeft, r, g, b))
			SimpleSelfRebuff:ScheduleEvent('SimpleSelfRebuff_UpdateDisplay', 1)
			--self:SetIcon(BS:GetSpellIcon(buffToCast))
		else
			self:SetText('SSR')
			--self:SetIcon(false)
		end
	else
		self:SetText('|cff777777SSR|r')
		--self:SetIcon(false)
	end
end

function SimpleSelfRebuff:OnTooltipUpdate()
	local cat
	Tablet:SetTitle(self:GetTitle())
	if not active then
		cat = Tablet:AddCategory()
		cat:AddLine('text', L['Currently inactive (stealthed/in combat/dead/mounted/resting).'])
	else
		cat = Tablet:AddCategory("columns", 3)
		for categoryName, category in pairs(categories) do
			if category.count > 0 then
				local name, timeLeft, r, g, b = GetBuffInfo(category)
				cat:AddLine(
					'text',  categoryName,
					'text2', name,     'text2R', r, 'text2G', g, 'text2B', b,
					'text3', timeLeft, 'text3R', r, 'text3G', g, 'text3B', b
				)
			end
		end
		SimpleSelfRebuff:ScheduleEvent('SimpleSelfRebuff_UpdateDisplay', 1)		
		Tablet:SetHint(L['Green means "ok", yellow means "override wanted buff", red means "missing".'])
	end
end


